@@ -1,10 +1,12 @@
-//! Scriptable SSH through OpenSSH.
+//! Scriptable SSH through OpenSSH (**only works on unix**).
 //!
 //! This crate wraps the OpenSSH remote login client (`ssh` on most machines), and provides
 //! a convenient mechanism for running commands on remote hosts. Since all commands are executed
 //! through the `ssh` command, all your existing configuration (e.g., in `.ssh/config`) should
 //! continue to work as expected.
 //!
+//! # Executing remote processes
+//!
 //! The library's API is modeled closely after that of [`std::process::Command`], since `ssh` also
 //! attempts to make the remote process seem as much as possible like a local command. However,
 //! there are some differences.
@@ -12,10 +14,6 @@
 //! First of all, all remote commands are executed in the context of a single ssh
 //! [session](Session). Authentication happens once when the session is
 //! [established](Session::connect), and subsequent command invocations re-use the same connection.
-//! Behind the scenes, the crate uses ssh's [`ControlMaster`] feature to multiplex the channels for
-//! the different remote commands. Because of this, each remote command is tied to the lifetime of
-//! the [`Session`] that spawned them. When the session is [closed](Session::close), the connection
-//! is severed, and there can be no outstanding remote clients.
 //!
 //! Note that the maximum number of multiplexed remote commands is 10 by default. This value can be
 //! increased by changing the `MaxSessions` setting in [`sshd_config`].
@@ -35,9 +33,41 @@
 //! the remote command directly. Usually, these are the same, though not always, as highlighted in
 //! the documetantation the individual methods. See also the section below on Remote Shells.
 //!
-//! And finally, our commands never default to inheriting stdin/stdout/stderr, since we expect you
-//! are using this to automate things. Instead, unless otherwise noted, all I/O ports default to
-//! [`Stdio::null`](std::process::Stdio::null).
+//! # Connection modes
+//!
+//! This library provides two way to connect to the [`ControlMaster`]:
+//!
+//! One is to spawn a new process, the other is to connect to
+//! the control socket directly.
+//!
+//! The process implementation executes remote commands by invoking
+//! the ssh command locally with arguments that make the invocation
+//! reuse the connections set up by the control master.
+//!
+//! This maximizes compatibility with OpenSSH, but loses out on some fidelity
+//! in information about execution since only the exit code and the output of
+//! the ssh command is available to inspect.
+//!
+//! The native mux implementation on the other hand connects directly to
+//! the ssh control master and executes commands and retrieves the exit codes and
+//! the output of the remote process over its native protocol.
+//!
+//! This gives better access to error information at the cost of introducing
+//! more non-OpenSSH code into the call path.
+//!
+//! The former parses the stdout/stderr of the ssh control master to retrieve the error
+//! for any failed operations, while the later retrieves the error from the control socket
+//! directly.
+//!
+//! Thus, the error handling in the later is more robust.
+//!
+//! Also, the former requires one process to be spawn for every connection while the later only
+//! needs to create one socket, so the later has better performance and consumes less resource.
+//!
+//! Behind the scenes, the crate uses ssh's [`ControlMaster`] feature to multiplex the channels for
+//! the different remote commands. Because of this, each remote command is tied to the lifetime of
+//! the [`Session`] that spawned them. When the session is [closed](Session::close), the connection
+//! is severed, and there can be no outstanding remote clients.
 //!
 //! # Authentication
 //!
@@ -84,11 +114,13 @@
 //! # Examples
 //!
 //! ```rust,no_run
+//! # #[cfg(feature = "native-mux")]
 //! # #[tokio::main]
 //! # async fn main() -> Result<(), openssh::Error> {
 //! use openssh::{Session, KnownHosts};
 //!
-//! let session = Session::connect("me@ssh.example.com", KnownHosts::Strict).await?;
+//! let session = Session::connect_mux("me@ssh.example.com", KnownHosts::Strict).await?;
+//!
 //! let ls = session.command("ls").output().await?;
 //! eprintln!("{}", String::from_utf8(ls.stdout).expect("server output was not valid UTF-8"));
 //!
@@ -106,16 +138,27 @@
 #![warn(
     missing_docs,
     missing_debug_implementations,
-    broken_intra_doc_links,
+    rustdoc::broken_intra_doc_links,
     rust_2018_idioms,
     unreachable_pub
 )]
+#![cfg_attr(
+    not(any(feature = "process-mux", feature = "native-mux")),
+    allow(unused_variables, unreachable_code, unused_imports, dead_code)
+)]
+// only enables the nightly `doc_cfg` feature when
+// the `docsrs` configuration attribute is defined
+#![cfg_attr(docsrs, feature(doc_cfg))]
+
+#[cfg(not(unix))]
+compile_error!("This crate can only be used on unix");
 
 use std::borrow::Cow;
 use std::ffi::OsStr;
-use std::io;
-use tokio::io::AsyncReadExt;
-use tokio::process;
+use std::path::Path;
+
+mod stdio;
+pub use stdio::{ChildStderr, ChildStdin, ChildStdout, Stdio};
 
 mod builder;
 pub use builder::{KnownHosts, SessionBuilder};
@@ -129,32 +172,81 @@ pub use child::RemoteChild;
 mod error;
 pub use error::Error;
 
-mod sftp;
-pub use sftp::{Mode, RemoteFile, Sftp};
+#[cfg(feature = "process-mux")]
+pub(crate) mod process_impl;
+
+#[cfg(feature = "native-mux")]
+pub(crate) mod native_mux_impl;
+
+#[cfg(doc)]
+/// Changelog for this crate.
+pub mod changelog;
+
+mod port_forwarding;
+pub use port_forwarding::*;
+
+/// Types to create and interact with the Remote Process
+pub mod process {
+    pub use super::{ChildStderr, ChildStdin, ChildStdout, Command, RemoteChild, Stdio};
+}
+
+#[derive(Debug)]
+pub(crate) enum SessionImp {
+    #[cfg(feature = "process-mux")]
+    ProcessImpl(process_impl::Session),
+
+    #[cfg(feature = "native-mux")]
+    NativeMuxImpl(native_mux_impl::Session),
+}
+
+#[cfg(any(feature = "process-mux", feature = "native-mux"))]
+macro_rules! delegate {
+    ($impl:expr, $var:ident, $then:block) => {{
+        match $impl {
+            #[cfg(feature = "process-mux")]
+            SessionImp::ProcessImpl($var) => $then,
+
+            #[cfg(feature = "native-mux")]
+            SessionImp::NativeMuxImpl($var) => $then,
+        }
+    }};
+}
+
+#[cfg(not(any(feature = "process-mux", feature = "native-mux")))]
+macro_rules! delegate {
+    ($impl:expr, $var:ident, $then:block) => {{
+        unreachable!("Neither feature process-mux nor native-mux is enabled")
+    }};
+}
 
 /// A single SSH session to a remote host.
 ///
-/// You can use [`command`] to start a new command on the connected machine.
+/// You can use [`command`](Session::command) to start a new command on the connected machine.
 ///
 /// When the `Session` is dropped, the connection to the remote host is severed, and any errors
 /// silently ignored. To disconnect and be alerted to errors, use [`close`](Session::close).
 #[derive(Debug)]
-pub struct Session {
-    ctl: tempfile::TempDir,
-    addr: String,
-    terminated: bool,
-    master: std::sync::Mutex<Option<(tokio::process::ChildStdout, tokio::process::ChildStderr)>>,
+pub struct Session(SessionImp);
+
+#[cfg(feature = "process-mux")]
+impl From<process_impl::Session> for Session {
+    fn from(imp: process_impl::Session) -> Self {
+        Self(SessionImp::ProcessImpl(imp))
+    }
+}
+
+#[cfg(feature = "native-mux")]
+impl From<native_mux_impl::Session> for Session {
+    fn from(imp: native_mux_impl::Session) -> Self {
+        Self(SessionImp::NativeMuxImpl(imp))
+    }
 }
 
 // TODO: UserKnownHostsFile for custom known host fingerprint.
-// TODO: Extract process output in Session::check(), Session::connect(), and Session::terminate().
 
 impl Session {
-    fn ctl_path(&self) -> std::path::PathBuf {
-        self.ctl.path().join("master")
-    }
-
-    /// Connect to the host at the given `addr` over SSH.
+    /// Connect to the host at the given `host` over SSH using process impl, which will
+    /// spawn a new ssh process for each `Child` created.
     ///
     /// The format of `destination` is the same as the `destination` argument to `ssh`. It may be
     /// specified as either `[user@]hostname` or a URI of the form `ssh://[user@]hostname[:port]`.
@@ -164,42 +256,50 @@ impl Session {
     /// instead.
     ///
     /// For more options, see [`SessionBuilder`].
+    #[cfg(feature = "process-mux")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "process-mux")))]
     pub async fn connect<S: AsRef<str>>(destination: S, check: KnownHosts) -> Result<Self, Error> {
         let mut s = SessionBuilder::default();
         s.known_hosts_check(check);
         s.connect(destination.as_ref()).await
     }
 
-    /// Check the status of the underlying SSH connection.
+    /// Connect to the host at the given `host` over SSH using native mux impl, which
+    /// will create a new socket connection for each `Child` created.
     ///
-    /// Since this does not run a remote command, it has a better chance of extracting useful error
-    /// messages than other commands.
+    /// See the crate-level documentation for more details on the difference between native and process-based mux.
+    ///
+    /// The format of `destination` is the same as the `destination` argument to `ssh`. It may be
+    /// specified as either `[user@]hostname` or a URI of the form `ssh://[user@]hostname[:port]`.
+    ///
+    /// If connecting requires interactive authentication based on `STDIN` (such as reading a
+    /// password), the connection will fail. Consider setting up keypair-based authentication
+    /// instead.
+    ///
+    /// For more options, see [`SessionBuilder`].
+    #[cfg(feature = "native-mux")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "native-mux")))]
+    pub async fn connect_mux<S: AsRef<str>>(
+        destination: S,
+        check: KnownHosts,
+    ) -> Result<Self, Error> {
+        let mut s = SessionBuilder::default();
+        s.known_hosts_check(check);
+        s.connect_mux(destination.as_ref()).await
+    }
+
+    /// Check the status of the underlying SSH connection.
+    #[cfg(not(windows))]
+    #[cfg_attr(docsrs, doc(cfg(not(windows))))]
     pub async fn check(&self) -> Result<(), Error> {
-        if self.terminated {
-            return Err(Error::Disconnected);
-        }
+        delegate!(&self.0, imp, { imp.check().await })
+    }
 
-        let check = process::Command::new("ssh")
-            .arg("-S")
-            .arg(self.ctl_path())
-            .arg("-o")
-            .arg("BatchMode=yes")
-            .arg("-O")
-            .arg("check")
-            .arg(&self.addr)
-            .output()
-            .await
-            .map_err(Error::Ssh)?;
-
-        if let Some(255) = check.status.code() {
-            if let Some(master_error) = self.take_master_error().await {
-                Err(master_error)
-            } else {
-                Err(Error::Disconnected)
-            }
-        } else {
-            Ok(())
-        }
+    /// Get the SSH connection's control socket path.
+    #[cfg(not(windows))]
+    #[cfg_attr(docsrs, doc(cfg(not(windows))))]
+    pub fn control_socket(&self) -> &Path {
+        delegate!(&self.0, imp, { imp.ctl() })
     }
 
     /// Constructs a new [`Command`] for launching the program at path `program` on the remote
@@ -207,7 +307,7 @@ impl Session {
     ///
     /// Before it is passed to the remote host, `program` is escaped so that special characters
     /// aren't evaluated by the remote shell. If you do not want this behavior, use
-    /// [`raw_command`].
+    /// [`raw_command`](Session::raw_command).
     ///
     /// The returned `Command` is a builder, with the following default configuration:
     ///
@@ -226,7 +326,7 @@ impl Session {
     /// Constructs a new [`Command`] for launching the program at path `program` on the remote
     /// host.
     ///
-    /// Unlike [`command`], this method does not shell-escape `program`, so it may be evaluated in
+    /// Unlike [`command`](Session::command), this method does not shell-escape `program`, so it may be evaluated in
     /// unforeseen ways by the remote shell.
     ///
     /// The returned `Command` is a builder, with the following default configuration:
@@ -240,24 +340,10 @@ impl Session {
     /// If `program` is not an absolute path, the `PATH` will be searched in an OS-defined way on
     /// the host.
     pub fn raw_command<S: AsRef<OsStr>>(&self, program: S) -> Command<'_> {
-        // XXX: Should we do a self.check() here first?
-
-        // NOTE: we pass -p 9 nine here (the "discard" port) to ensure that ssh does not
-        // succeed in establishing a _new_ connection if the master connection has failed.
-
-        let mut cmd = process::Command::new("ssh");
-        cmd.arg("-S")
-            .arg(self.ctl_path())
-            .arg("-T")
-            .arg("-o")
-            .arg("BatchMode=yes")
-            .arg("-p")
-            .arg("9")
-            .arg(&self.addr)
-            .arg("--")
-            .arg(program);
-
-        Command::new(self, cmd)
+        Command::new(
+            self,
+            delegate!(&self.0, imp, { imp.raw_command(program).into() }),
+        )
     }
 
     /// Constructs a new [`Command`] that runs the provided shell command on the remote host.
@@ -265,10 +351,10 @@ impl Session {
     /// The provided command is passed as a single, escaped argument to `sh -c`, and from that
     /// point forward the behavior is up to `sh`. Since this executes a shell command, keep in mind
     /// that you are subject to the shell's rules around argument parsing, such as whitespace
-    /// splitting, variable expansion, and other funkyness. I _highly_ recommend you read [this
-    /// article] if you observe strange things.
+    /// splitting, variable expansion, and other funkyness. I _highly_ recommend you read
+    /// [this article] if you observe strange things.
     ///
-    /// While the returned `Command` is a builder, like for [`command`], you should not add
+    /// While the returned `Command` is a builder, like for [`command`](Session::command), you should not add
     /// additional arguments to it, since the arguments are already passed within the shell
     /// command.
     ///
@@ -293,8 +379,8 @@ impl Session {
     /// With sufficiently complex or weird commands, the escaping of `shell-escape` may not fully
     /// match the "un-escaping" of the remote shell. This will manifest as escape characters
     /// appearing in the `sh` command that you did not intend to be there. If this happens, try
-    /// changing the remote shell if you can, or fall back to [`command`] and do the escaping
-    /// manually instead.
+    /// changing the remote shell if you can, or fall back to [`command`](Session::command)
+    /// and do the escaping manually instead.
     ///
     ///   [POSIX compliant]: https://pubs.opengroup.org/onlinepubs/9699919799/xrat/V4_xcu_chap02.html
     ///   [this article]: https://mywiki.wooledge.org/Arguments
@@ -305,101 +391,31 @@ impl Session {
         cmd
     }
 
-    /// Prepare to perform file operations on the remote host.
+    /// Request to open a local/remote port forwarding.
+    /// The `Socket` can be either a unix socket or a tcp socket.
     ///
-    /// See [`Sftp`] for details on how to interact with the remote files.
-    pub fn sftp(&self) -> Sftp<'_> {
-        Sftp::new(self)
+    /// If `forward_type` == Local, then `listen_socket` on local machine will be
+    /// forwarded to `connect_socket` on remote machine.
+    ///
+    /// Otherwise, `listen_socket` on the remote machine will be forwarded to `connect_socket`
+    /// on the local machine.
+    ///
+    /// Currently, there is no way of stopping a port forwarding due to the fact that
+    /// openssh multiplex server/master does not support this.
+    pub async fn request_port_forward(
+        &self,
+        forward_type: ForwardType,
+        listen_socket: Socket<'_>,
+        connect_socket: Socket<'_>,
+    ) -> Result<(), Error> {
+        delegate!(&self.0, imp, {
+            imp.request_port_forward(forward_type, listen_socket, connect_socket)
+                .await
+        })
     }
 
     /// Terminate the remote connection.
-    pub async fn close(mut self) -> Result<(), Error> {
-        self.terminate().await
-    }
-
-    async fn take_master_error(&self) -> Option<Error> {
-        let (_stdout, mut stderr) = self.master.lock().unwrap().take()?;
-
-        let mut err = String::new();
-        if let Err(e) = stderr.read_to_string(&mut err).await {
-            return Some(Error::Master(e));
-        }
-        let stderr = err.trim();
-
-        if stderr.is_empty() {
-            return None;
-        }
-
-        let kind = if stderr.contains("Connection to") && stderr.contains("closed by remote host") {
-            io::ErrorKind::ConnectionAborted
-        } else {
-            io::ErrorKind::Other
-        };
-
-        Some(Error::Master(io::Error::new(kind, stderr)))
-    }
-
-    async fn terminate(&mut self) -> Result<(), Error> {
-        if !self.terminated {
-            let exit = process::Command::new("ssh")
-                .arg("-S")
-                .arg(self.ctl_path())
-                .arg("-o")
-                .arg("BatchMode=yes")
-                .arg("-O")
-                .arg("exit")
-                .arg(&self.addr)
-                .output()
-                .await
-                .map_err(Error::Ssh)?;
-
-            self.terminated = true;
-
-            if let Some(master_error) = self.take_master_error().await {
-                return Err(master_error);
-            }
-
-            if exit.status.success() {
-                return Ok(());
-            }
-
-            // let's get this case straight:
-            // we tried to tell the master to exit.
-            // the -O exit command failed.
-            // the master exited, but did not produce an error.
-            // what could cause that?
-            //
-            // the only thing I can think of at the moment is that the remote end cleanly
-            // closed the connection, probably by virtue of being killed (but without the
-            // network dropping out). since we were told to _close_ the connection, well, we
-            // have succeeded, so this should not produce an error.
-            //
-            // we will still _collect_ the error that -O exit produced though,
-            // just for ease of debugging.
-
-            let _exit_err = String::from_utf8_lossy(&exit.stderr);
-            let _err = _exit_err.trim();
-            // eprintln!("{}", _err);
-        }
-
-        Ok(())
-    }
-}
-
-impl Drop for Session {
-    fn drop(&mut self) {
-        if !self.terminated {
-            let _ = std::process::Command::new("ssh")
-                .arg("-S")
-                .arg(self.ctl_path())
-                .arg("-o")
-                .arg("BatchMode=yes")
-                .arg("-O")
-                .arg("exit")
-                .arg(&self.addr)
-                .stdout(std::process::Stdio::null())
-                .stderr(std::process::Stdio::null())
-                .status();
-        }
+    pub async fn close(self) -> Result<(), Error> {
+        delegate!(self.0, imp, { imp.close().await })
     }
 }
